#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Stand: Mittwoch 23.01.19

"""

# In diesem Modul finden Sie die aktuelle Musterlösung zur Polynomklasse und zu
# verschiedenen Aufgaben, die mit dieser Klasse zusammenhängen, u.a.:
# Musterlösung von Blatt 8, Aufgabe 31
# Musterlösung von Blatt 10, Aufgabe 39
# Musterlösung von Blatt 11, Aufgabe 42

# Sie können die Datei als Skript ausführen, oder als Modul z.B. mit
# import polynompy as pp 
# laden.

# Neue Ergänzungen/Erweiterungen in der Klasse Polynom, die weder im Skript 
# stehen, noch explizit in einer Aufgabe gefordert wurden:
# - __add__ und __sub__ gekürzt
# - skalare Addition und Subtraktion ergänzt (siehe __add__, __sub__, __radd__, __rsub__)
# - skalare Division ergänzt (siehe __truediv__)
# - Plot-Routine erweitert, sodass nun eine Figure und ein Axen-Objekt ausgegeben werden 
# - __repr__ stellt nun das Polynom mit gekürtzten Koeffizienten dar

class Polynom():
    """ Polynomklasse """
    # Hier ist es sinnvoll, die benötigten Pakete direkt in die Klasse zu laden.
    # So ist sichergestellt, dass die Klasse die notwendigen
    # Pakete immer kennt.
    # Methoden aus diesen Pakten müssen dann mit einem "self" davor geladen werden.
    import numpy as np
    import matplotlib.pyplot as plt
    
    def __init__(self, dp):
        # falls die Funktion konstant ist
        if isinstance(dp, (int, float, complex)):
            dp = {0:dp}
        self.dp = dp
        self.degree = max(dp.keys())
        assert all(isinstance(x, int) and x >= 0 for x in dp.keys()),\
            "Die Exponenten (keys) müssen natürliche Zahlen (einschl. 0) sein."
        
    def __repr__(self):
        polystr = ''
        for k in sorted(self.dp):
            polystr = polystr + '+{0:.2f}*X^{1}'.format(self.dp[k], k)
        polystr = polystr.replace('X^0', '1')
        polystr = polystr.replace('+-', '-')
        polystr = polystr.replace('1*', '')
        polystr = polystr.replace('*1', '')
        polystr = polystr.replace('X^1', 'X')
        if polystr[0] == '+':
            polystr = polystr[1:]
        return 'Polynom: ' + polystr 
    
    def __add__(self, other):
        # Skalare Addition
        if isinstance(other, (int, float, complex)):
            other = Polynom(other)
        # eine Möglichkeit zur Addition:
#        erg = self.dp.copy()
#        for k in other.dp:
#            if k in self.dp:
#                erg[k] += other.dp[k]
#            else:
#                erg[k] = other.dp[k]
        # hier ist eine noch kürzere Variante: .get(n,0) greift auf den Eintrag n 
        # des Dictionaries und gibt falls der nicht vorhanden ist
        # das zweite Argument (hier 0) zurück
        erg = dict((n, self.dp.get(n, 0) + other.dp.get(n, 0)) for n in set(self.dp).union(set(other.dp)))
        return Polynom(erg)
    
    def __sub__(self, other):
        # s.o.
        if isinstance(other, (int, float, complex)):
            other = Polynom(other)
#        erg = self.dp.copy()
#        for k in other.dp:
#            if k in self.dp:
#                erg[k] -= other.dp[k]
#            else:
#                erg[k] = -other.dp[k]
        erg = dict((n, self.dp.get(n, 0)-other.dp.get(n, 0)) for n in set(self.dp).union(set(other.dp)))
        return Polynom(erg)
    
    def __rsub__(self, other):
        # damit auch für ein Skalar a und ein Polynom p auch a-p funktioniert
        if isinstance(other, (int, float, complex)):
            other = Polynom(other)
        return other - self

    # Zum Auswerten der Funktion (normale Auswertung und Auswertung mit Horner ist möglich):
    def __call__(self, x, method=None):
        if method == None:
            return sum([self.dp[k]*x**k for k in self.dp])
        elif method == 'Horner':
            return self.__Horner__(x)
        
    # Multiplikation mit Skalaren und Polynomen:
    def __mul__(self, other):
        erg = dict()
        if isinstance(other, (int, float, complex)):
            for k in self.dp:
                erg[k] = self.dp[k] * other
        else:
            for i in self.dp:
                for j in other.dp:
                    if i+j in erg:
                        erg[i+j] += self.dp[i]*other.dp[j]
                    else:
                        erg[i+j] = self.dp[i]*other.dp[j]
        return Polynom(erg)
    
    def __truediv__(self, other):
        # Wir wollen für ein Polynom p und ein Skalar a p/2 fuer 1/2 * p schreiben können
        if isinstance(other, (int, float, complex)):
            erg = dict((n, self.dp[n]/other) for n in self.dp)
        else:
            raise NotImplementedError("Polynomdivision nicht implementiert")   
        return Polynom(erg)
    
    # Umgekehrte Multiplikation, so dass auch Skalar*Polynom funktioniert:        
    __rmul__ = __mul__
    # Umgekehrte Addition, so dass auch Skalar+Polynom funktioniert:        
    __radd__ = __add__
    
    def conjugate(self):
        """ Komplexe Konjugation der Koeffizienten """
        erg = self.dp.copy()
        for i in self.dp:
            if isinstance(self.dp[i], complex):
                erg[i] = self.dp[i].conjugate()
        return Polynom(erg)
    
    def diff(self):
        """ Ableitung des Polynoms """
        erg = dict()
        if self.degree == 0:
            erg[0] = 0
        else:
            for i in self.dp:
                if i != 0:
                    erg[i-1] = self.dp[i]*i
        return Polynom(erg)
    
    def integrate(self):
        """ Stammfunktion des Polynoms """
        erg = dict()
        for i in self.dp:
            erg[i+1] = self.dp[i]/(i+1)
        return Polynom(erg)
    
    def integral(self, a, b):
        """ Berechent Integral von a bis b """ 
        intPol = self.integrate()
        return intPol(b) - intPol(a)
    
    def plot(self, fig=None, ax=None, xmin=0, xmax=1, num=50, style='k--s',\
             method=None):
        """ Plotroutine für Klasse Polynom
        
         fig -- Figurehandle von pyplot.figure() \n
         ax -- Koordinatensysterm \n
         [xmin,xmax]  -- Intervall über dem gezeichnet werden soll \n
         num -- Anzahl der Punkte \n
         style -- Linienstil
        """
        if fig == None:
            fig = self.plt.figure()
            
        if ax == None:
            ax = fig.add_subplot(111)
        x = self.np.linspace(xmin, xmax, num)
        if method is None:
            ax.plot(x, self(x), style, label='${}$'.format(self))
        elif method == 'Horner':
            ax.plot(x, self.__Horner__(x), style, label='${}$'.format(self))
        else:
            raise NotImplementedError("Verfahren zur Auswertung nicht implementiert")
        ax.legend()
        return fig, ax
    
    def __Horner__(self, x):
        """ Auswertung mit dem Horner-Schema (Aufg. 33) """
        # Die Funktion ist "versteckt", d.h. p.Horner(x) ist nicht bekannt,
        # dafür aber p.__Horner__(x)
        erg = 0
        mk = self.degree
        for k in sorted(self.dp, reverse=True):
            erg = self.dp[k] + erg*x**(mk - k)
            mk = k
        return erg

def Monombasis(n):
    """Berechnet die Monombasis [1,x,...,x**n]"""
    return [Polynom({k:1}) for k in range(n+1)] 
    
def SProdKoeff(p, q):
    """Skalarprodukt über die Koeffizienten """
    from numpy import sqrt
    q = q.conjugate()
    return sqrt(sum([p.dp.get(n, 0)*q.conjugate().dp.get(n, 0) for n in set(p.dp).union(set(q.dp))]))

def SProdL2(p, q):
    """Skalarprodukt über das Integral von -1 bis 1"""
    return (p*q).integral(-1, 1)

# Musterlösung von Aufgabe 31 (unter Verwendung der Polynom-Klasse)
def Legendre(n, method=None):
    """ Legendre Polynom vom Grad n """
    assert n >= 0 and isinstance(n, int), \
        "n ist keine natürliche Zahl."
        
    if method == None: #iterative Version
        if n == 0: 
            return Polynom(1)
    
        P_nm1 = Polynom(1)
        P_n = Polynom({1:1})
        x = Polynom({1:1})
        for k in range(2, n+1):
            P_n, P_nm1 = ((2*k-1)*x*P_n - (k-1)*P_nm1) / k, P_n
        return P_n
    
    elif method == 'Rec': #rekursive Version
        if n == 0: 
            return Polynom(1)
        elif n == 1:
            return Polynom({1:1})
        else:
            x = Polynom({1:1})
        return ((2*n-1)*x*Legendre(n-1, method='Rec') - (n-1)*Legendre(n-2, method='Rec')) / n

# Musterlösung von Aufgabe 39 
def Lagrange(knots):
    import numpy as np
    LMat = np.zeros((len(knots), len(knots)), dtype = Polynom)
    i = 0
    for ci in sorted(knots):
        j = 0
        for cj in sorted(knots):
            if ci==cj:
                LMat[i,j] = Polynom({0:1})
            else:
                LMat[i,j] = 1/(ci-cj) * Polynom({0:-cj, 1:1})
            j += 1
        i += 1
    return list(np.prod(LMat, axis = 1))    

# Musterlösung von Aufgabe 42
def GramSchmidt(poly_list):
    n = len(poly_list)
    R = np.zeros((n, n))
    Q = []
    for k in range(n):
        Q.append(poly_list[k])
        if k != 0:
            for j in range (k):
                R[j, k] = SProdL2(Q[j], Q[k])
            for i in range(k):
                Q[k] -= Q[i] * R[i, k]
        R[k, k] = np.sqrt(SProdL2(Q[k], Q[k]))
        Q[k] /= R[k, k]
    return Q

# Wenn wir polynompy as Skript ausführen, werden die unten stehend Test ausgeführt,
# als import... nicht    
if __name__ == '__main__':
    # Test der Klasse Polynom
    import matplotlib.pyplot as plt
    import numpy as np
    
    p = Polynom({7:1/2, 6:-7/2, 5:21/2, 4:-35/2, 3:35/2, 2:-21/2, 1:7/2, 0:-1/2})
    q = Polynom({0:2, 2:3, 3:4, 6:-8.})

    print('q ', q, '2.*q ', 2.*q, 'q*2 ', q*2)
    print('q ', q, 'q/2. ', q/2., 'q/2 ', q/2)
    print('q+1. ', q+1., '1+q ', 1+q, 'q-1. ', q-1., '1-q ', 1-q)
    
    print(p(1))
    print(p(1, method='Horner'))
    print(p.__Horner__(1))
    
    fig = plt.figure(50)
    ax1 = fig.add_subplot(121)
    p.plot(fig, ax=ax1, xmin=0.99, xmax=1.01, style='b')
    p.plot(fig, ax=ax1, xmin=0.99, xmax=1.01, style='r', method='Horner')
    
    ax2 = fig.add_subplot(122)
    q.plot(fig, ax=ax2, xmin=0, xmax=1, style='b')
    q.plot(fig, ax=ax2, xmin=0, xmax=1, style='r', method='Horner')
    
    # Test von Legendre
    fig = plt.figure(100)
    fig, ax = Legendre(3, method='Rec').plot(fig)
    fig, ax = Legendre(2, method='Rec').plot(fig, ax)
    
    # Test von Lagrange
    cset = set(np.cos((2*np.arange(7)+1)*np.pi/14))
    LagPoly = Lagrange(cset)
    colors = ['r', 'g', 'b', 'y', 'c', 'm', 'k']
    j = 0
    fig = plt.figure(figsize=(10,7))
    ax = fig.add_subplot(111)
    for k in LagPoly:
        k.plot(fig = fig, ax = ax, xmin = -1, xmax = 1, style = colors[j]+'-', num = 200)
        j += 1